/************************************************************************
 *   IRC - Internet Relay Chat, src/gzlines.c
 *   Copyright (C) 1999 Maarten van den Bosch
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************************************************************/

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>

#include "h.h"


static int advanced_check(char *, int);

aGZline *GZlineList = NULL;

/* ok, given a mask, our job is to determine
 * wether or not it's a safe mask to banish...
 *
 * userhost= mask to verify
 * ipstat= TRUE  == it's an ip
 *         FALSE == it's a hostname
 *         UNSURE == we need to find out
 * return value
 *         TRUE  == mask is ok
 *         FALSE == mask is not ok
 *        UNSURE == [unused] something went wrong
 */

int advanced_check(char *userhost, int ipstat)
{
    register int retval = TRUE;
    char *up = NULL, *p, *thisseg;
    int numdots = 0, segno = 0, numseg, i = 0;
    char *ipseg[10 + 2];
    char safebuffer[512] = "";	/* buffer strtoken() can mess up to its heart's content...;> */

    strcpy(safebuffer, userhost);

#define userhost safebuffer
#define IP_WILDS_OK(x) ((x)<2? 0 : 1)

    if (ipstat == UNSURE) {
	ipstat = TRUE;
	for (; *up; up++) {
	    if (*up == '.')
		numdots++;
	    if (!isdigit(*up) && !ispunct(*up)) {
		ipstat = FALSE;
		continue;
	    }
	}
	if (numdots != 3)
	    ipstat = FALSE;
	if (numdots < 1 || numdots > 9)
	    return (0);
    }
    /* fill in the segment set */
    {
	int l = 0;
	for (segno = 0, i = 0, thisseg = strtoken(&p, userhost, "."); thisseg;
	     thisseg = strtoken(&p, NULL, "."), i++) {

	    l = strlen(thisseg) + 2;
	    ipseg[segno] = calloc(1, l);
	    strncpy(ipseg[segno++], thisseg, l);
	}
    }
    if (segno < 2 && ipstat == TRUE)
	retval = FALSE;
    numseg = segno;
    if (ipstat == TRUE)
	for (i = 0; i < numseg; i++) {
	    if (!IP_WILDS_OK(i) && (index(ipseg[i], '*') || index(ipseg[i], '?')))
		retval = FALSE;
	    MyFree(ipseg[i]);
    } else {
	int wildsok = 0;

	for (i = 0; i < numseg; i++) {
	    /* for hosts, let the mask extent all the way to
	       the second-level domain... */
	    wildsok = 1;
	    if (i == numseg || (i + 1) == numseg)
		wildsok = 0;
	    if (wildsok == 0 && (index(ipseg[i], '*') || index(ipseg[i], '?'))) {
		retval = FALSE;
	    }
	    MyFree(ipseg[i]);
	}


    }

    return (retval);
#undef userhost
#undef IP_WILDS_OK

}

int DelGZline(char *userhost, char *username)
{
    Reg1 aGZline *gzptr;
    Reg2 aGZline *gzptr2;

    if (GZlineList == NULL)
	return 0;

    for (gzptr = GZlineList; gzptr; gzptr = gzptr2) {

	gzptr2 = gzptr->next;

	if (!strcasecmp(gzptr->ip, userhost) && !match(gzptr->ident, username)) {
	    if (gzptr->prev)
		gzptr->prev->next = gzptr->next;
	    else
		GZlineList = gzptr->next;
	    if (gzptr->next)
		gzptr->next->prev = gzptr->prev;
	    free_gzline(gzptr);
	    return 1;
	}
    }

    return 0;
}

int GZlineExists(char *userhost, char *username)
{
    Reg1 aGZline *gzptr;
    Reg2 aGZline *gzptr2;

    if (GZlineList == NULL)
	return 0;

    for (gzptr = GZlineList; gzptr; gzptr = gzptr2) {

	gzptr2 = gzptr->next;

	if (!strcasecmp(gzptr->ip, userhost) && !strcasecmp(gzptr->ident, username)) {
	    return 1;
	}
    }

    return 0;
}

int find_gzlines(aClient * cptr)
{
    Reg1 aGZline *gzptr;
    Reg2 aGZline *gzptr2;

    char *username;

    username = cptr->username;

    if (GZlineList == NULL)
	return 0;

    for (gzptr = GZlineList; gzptr; gzptr = gzptr2) {

	gzptr2 = gzptr->next;

	if (!match(gzptr->ip, inetntoa((char *) &cptr->ip)) && (!match(gzptr->ident, username) || !strcasecmp(username, "unknown"))) {
	    sendto_one(cptr, "*** You are banned from this network by %s for %s.", gzptr->from, gzptr->reason);
	    sendto_one(cptr, "*** This ban will last until %s.", myctime(gzptr->duration));
	    sendto_one(cptr, "*** Please send an e-mail to kline@starchat.net for more information.");
	    sendto_ops("GZ-line active for %s!%s@%s", cptr->name, cptr->username, cptr->user->host);
	    return 1;
	}
    }

    return 0;
}

void AddGZline(char *ip, char *ident, char *reason, char *from, int flag, int duration, int timeset)
{
    Reg1 aGZline *gzptr;

    gzptr = make_gzline();

    strncpyzt(gzptr->ip, ip, sizeof(gzptr->ip));
    strncpyzt(gzptr->ident, ident, sizeof(gzptr->ident));
    strncpyzt(gzptr->reason, reason, sizeof(gzptr->reason));
    strncpyzt(gzptr->from, from, sizeof(gzptr->from));
    gzptr->flag = flag;
    gzptr->duration = duration;
    gzptr->timeset = timeset;

    if (GZlineList == NULL)
	GZlineList = gzptr;
    else {
	GZlineList->prev = gzptr;
	gzptr->next = GZlineList;
	GZlineList = gzptr;
    }

    return;
}

/*
   ** m_gzline
   **      parv[0] = sender prefix
   **      parv[1] = mask ([ident]@ip)
   **      parv[2] = reason
   **      parv[3] = duration
   **
 */
int m_gzline(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int parc;
    char *parv[];
{
    aClient *acptr;
    char userhost[512 + 2] = "", *in;
    char ident[20] = "";
    char *mask, *reason, *from;
    int now, timeset, duration, i, j;
    time_t t1;

    mask = reason = from = NULL;

    now = time(&t1);

    /* This part handles GZ-lines from servers */

    if (IsServer(sptr)) {
	return 0;
    }
    /* This part handles GZ-lines from IRC Operators */

    if (!IsSAdmin(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return -1;
    }

    if (parc < 4) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "GZLINE");
	return -1;
    }

    if(find_services()) {
	sendto_one(sptr, err_str(ERR_SERVICESUP), me.name, parv[0]);
        return -1;
    }

    timeset = now;
    from = parv[0];
    mask = parv[1];
    reason = parv[2];
    duration = now + atoi(parv[3]);

    /* Check for ident and valid IP address */

    if ((in = index(parv[1], '@')) && (*(in + 1) != '\0')) {
	strcpy(userhost, in + 1);
	in = &userhost[0];
	while (*in) {
	    if (!isdigit(*in) && !ispunct(*in)) {
		sendto_one(sptr, ":%s NOTICE %s :GZ-lines work only with ip addresses.", me.name, sptr->name);
		return -1;
	    }
	    in++;
	}

	j = 0;

	while ((mask[j] != '@')) {
	    j++;
	}

	j++;

	strncpyzt(ident, mask, j);

    } else if (in && !(*(in + 1))) {
	sendto_one(sptr, ":%s NOTICE %s :Hello?! GZ-lines need an ip address!",
		   me.name, sptr->name);
	return -1;
    } else {
	strcpy(userhost, parv[1]);
	in = &userhost[0];
	while (*in) {
	    if (!isdigit(*in) && !ispunct(*in)) {
		sendto_one(sptr, ":%s NOTICE %s :GZ-lines work only with ip addresses.", me.name, sptr->name);
		return -1;
	    }
	    in++;
	    strncpyzt(ident, "*", 2);
	}
    }

    if (advanced_check(userhost, TRUE) == FALSE) {
	sendto_ops("Bad GZ-line mask from %s *@%s [%s]", parv[0], userhost, reason ? reason : "");
	if (MyClient(sptr))
	    sendto_one(sptr, ":%s NOTICE %s :*@%s is a bad GZ-line mask...", me.name, sptr->name, userhost);
	return -1;
    }
    if (GZlineExists(userhost, ident)) {
	if (MyClient(sptr))
	    sendto_one(sptr, ":%s NOTICE %s :*@%s already exists in my GZ-line list.", me.name, sptr->name, userhost);
	return -1;
    }
    if ((duration - timeset) < 3600)
	duration = timeset + 3600;

    sendto_ops("%s added GZ-line for %s@%s (%s) for %i seconds", from, ident, userhost, reason, (duration - timeset));

    AddGZline(userhost, ident, reason, from, 2, duration, timeset);

    for (i = highest_fd; i > 0; i--) {

	if ((acptr = local[i])) {

	    if (IsLog(acptr) || IsMe(acptr)) {
		continue;
	    }
	    if (find_gzlines(acptr)) {
		return exit_client(acptr, acptr, &me, "GZ-Lined");
	    }
	}
    }
   
    sendto_serv_butone(":%s GZLINE %s %s %s", me.name, parv[1], parv[2], parv[3]);

    return 0;
}

/*
   ** m_ungzline
   **      parv[0] = sender prefix
   **      parv[1] = mask ([ident]@ip)
 */
int m_ungzline(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int parc;
    char *parv[];
{

    char userhost[512 + 2] = "", *in;
    char ident[20] = "";
    char *mask, *from;
    int j;

    if (!IsSAdmin(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return -1;
    }
    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "GZLINE");
	return -1;
    }
    from = parv[0];
    mask = parv[1];

    /* Check for ident and valid IP address */

    if ((in = index(parv[1], '@')) && (*(in + 1) != '\0')) {
	strcpy(userhost, in + 1);
	in = &userhost[0];
	while (*in) {
	    if (!isdigit(*in) && !ispunct(*in)) {
		sendto_one(sptr, ":%s NOTICE %s :GZ-lines work only with ip addresses.", me.name, sptr->name);
		return -1;
	    }
	    in++;
	}

	j = 0;

	while ((mask[j] != '@')) {
	    j++;
	}

	j++;

	strncpyzt(ident, mask, j);

    } else if (in && !(*(in + 1))) {
	sendto_one(sptr, ":%s NOTICE %s :Hello?! GZ-lines need an ip address!",
		   me.name, sptr->name);
	return -1;
    } else {
	strcpy(userhost, parv[1]);
	in = &userhost[0];
	while (*in) {
	    if (!isdigit(*in) && !ispunct(*in)) {
		sendto_one(sptr, ":%s NOTICE %s :GZ-lines work only with ip addresses.", me.name, sptr->name);
		return -1;
	    }
	    in++;
	    strncpyzt(ident, "*", 2);
	}
    }

    if (!DelGZline(userhost, ident)) {
	if (MyClient(sptr))
	    sendto_one(sptr, ":%s NOTICE %s :*@%s does not exists in my GZ-line list.", me.name, sptr->name, userhost);
	return -1;
    }
    sendto_ops("%s deleted GZ-line for %s@%s", from, ident, userhost);

    return 1;
}
